# Vector Graphics Library
# vgllib.py
import pygame
import uuid
import time
import threading
import math

class Graph:
    @staticmethod
    def rect(frame, pos: tuple = (0, 0), size: tuple = (1, 1), color: tuple = (255, 255, 255), width: int = 0):
        pygame.draw.rect(frame.surface, color=color, rect=(pos[0], pos[1], size[0], size[1]), width=width)

    @staticmethod
    def line(frame, start_pos: tuple, end_pos: tuple, color: tuple = (255, 255, 255)):
        pygame.draw.aaline(frame.surface, color, start_pos, end_pos)

    @staticmethod
    def circle(frame, center: tuple, radius: int, color: tuple = (255, 255, 255), width: int = 0):
        pygame.draw.circle(frame.surface, color, center, radius, width)

    @staticmethod
    def ellipse(frame, pos: tuple = (0, 0), size: tuple = (1, 1), color: tuple = (255, 255, 255), width: int = 0):
        rect=(pos[0], pos[1], size[0], size[1])
        pygame.draw.ellipse(frame.surface, color, rect, width)

    @staticmethod
    def polygon(frame, pointlist: list, color: tuple = (255, 255, 255), width: int = 0):
        pygame.draw.polygon(frame.surface, color, pointlist, width)

    @staticmethod
    def arc(frame, pos: tuple = (0, 0), size: tuple = (1, 1), color: tuple = (255, 255, 255), start_angle: float = 0, stop_angle: float = 3.14, width: int = 1):
        rect=(pos[0], pos[1], size[0], size[1])
        pygame.draw.arc(frame.surface, color, rect, start_angle, stop_angle, width)

    @staticmethod
    def point(frame, pos: tuple, color: tuple = (255, 255, 255)):
        pygame.draw.point(frame.surface, color, pos)

    @staticmethod
    def lines(frame, pointlist: list, color: tuple = (255, 255, 255)):
        pygame.draw.aalines(frame.surface, color, closed=False, points=pointlist)

    @staticmethod
    def call(frame, method, **kwargs):
        if hasattr(Graph, method):
            getattr(Graph, method)(frame, **kwargs)
        else:
            print(f"方法 {method} 不存在")

class Frame(object):
    components = dict()
    components_stat = dict()
    render_thread = None
    motion_queue = list()
    def __init__(self, name: str, size: tuple):
        self.name = name
        self.size = size
        self.surface = pygame.Surface(size, flags=pygame.HWSURFACE)
        self.is_hide = False
        print("初始化子模块")
        self.render_thread = threading.Thread(target=self.render)
        print("启动图形渲染子线程")
        self.render_thread.start()

    def move(self, subname, direction, length, duration = 0, effect="linear"): # our powerful move!
        # direction: 使用角度制, 以直角笛卡尔坐标系的x正半轴方向为0度, 逆时针为加, 接受负数
        # length: 百分数
        # duration: "动画"时间, 为0则即时
        # effect: "动画"效果, linear为线性移动, 或许会在未来增加贝塞尔曲线
        # TODO: 增加 bezier 曲线
        if duration == 0:
            self.components[subname]["pos"][0] += math.cos(math.radians(direction)) * length / 100 * self.size[0]
            self.components[subname]["pos"][1] += math.sin(math.radians(direction)) * length / 100 * self.size[1]
            return
        self.motion_queue.append({"subname":subname, "direction":direction, "length":length, "start":round(time.time(), 1), "duration":duration, "effect":"linear"})
    
    def render(self):
        while 1:
            rest = list()
            while self.motion_queue:
                i = self.motion_queue.pop(0)  # 从队列的开头移除元素
                self.move(subname=i['subname'], direction=i['direction'], 
                        length=((time.time() - i['start']) / i['duration']) * i['length'], 
                        duration=0)
                if ((time.time() - i['start']) / i['duration']) > 0:
                    i['start'] = time.time()
                    i['duration'] -= (time.time() - i['start'])
                    rest.append(i)
            self.draw_all()
            self.motion_queue = rest
            time.sleep(0.1)
            

    def show(self, window, position: tuple):
        if not self.is_hide:
            window.blit(self.surface, position)

    def set_visible(self, newstat=True):
        self.is_hide = newstat

    def set_position(self, newposition):
        self.position = newposition

    def register(self, subname="", attr=None):
        if subname == "":
            subname = uuid.uuid4()
        # use percent of frame size instead of pixels :)
        attr['pos'] = list(attr['pos'])
        attr['pos'][0] = round(attr['pos'][0] / 100 * self.size[0])
        attr['pos'][1] = round(attr['pos'][1] / 100 * self.size[1])
        attr['size'] = list(attr['size'])
        attr['size'][0] = round(attr['size'][0] / 100 * self.size[0])
        attr['size'][1] = round(attr['size'][1] / 100 * self.size[1])
        self.components[subname] = attr
        self.components_stat[subname] = 1 # by default, not hiding

    def draw(self, attr):
        Graph.call(self, **attr)
    
    def set_component_visible(self, subname, newstat):
        self.components_stat[subname] = newstat

    def draw_all(self):
        for i in self.components.keys():
            if self.components_stat[i]:
                self.draw(self.components[i])
    
    def refresh(self, color=(0,0,0)):
        self.surface.fill(color)
    
    def clear(self):
        components = dict()
        components_stat = dict()

    def loads(self, grap_str):
        # TODO: 将会重写 以替代不安全的 eval
        self.register(subname="", attr=(eval(grap_str)))

    def load(self, file="default.vgld", mode="a"):
        # a: 增量加载 (默认)
        # w: 覆盖式加载
        # 文件扩展名: vgld (矢量图形层描述文件)
        if mode == 'w':
            self.clear()
        with open(file=file, mode="r+", encoding="UTF-8") as f:
            for i in f.readlines():
                #print(i)
                self.loads(i)


# 示例
if __name__ == "__main__":
    frame = None
    def grap():
        global frame
        pygame.init()
        window = pygame.display.set_mode((1200, 900))
        frame = Frame("Test", (1200, 900))
        #frame.load()
        frame.register(subname="test", attr={'method': 'rect', 'pos': (33, 33), 'size': (50, 50), 'color': (255, 255, 255)})
        #frame.move(subname='test', direction=0, length=90, duration=1, effect="linear")
        running = True
        while running:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    running = False

            window.fill((0, 0, 0))
            frame.show(window, (0, 0))
            pygame.display.flip()
            pygame.time.delay(10)
        pygame.quit()
    def debug():
        global frame
        while 1:
            try:
                e = input(">>>")
                if e == "f":
                    e = "frame.move(subname='test', direction=0, length=3, duration=1, effect='linear')"
                if e == "g":
                    e = "frame.move(subname='test', direction=180, length=3, duration=1, effect='linear')"
                exec(e)
            except:
                print("ER")
    grap_thd = threading.Thread(target=grap)
    debug_thd = threading.Thread(target=debug)
    grap_thd.start()
    debug_thd.start()